import base64
import json
import hashlib
import random
import time
import calendar
import requests
import hmac

from aes import AESCrypt
from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import pkcs1_15
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import serialization


user_id = '1' # '1'
url = 'http://service.local.test' # 'http://service.local.test'
safecode = '8b998abba49ae33a194cb9677b13ecbe' # '164ddce3d0b13d21fbc4cd02e14d9f98'
request_type = 'payment_v2' # payment_v2, withdraw_v2, payment_query_v2, withdraw_query_v2, balance_v2
transaction_secret = 'f55c19747fde060a4af40dc6ed187ae3'
guid = 'smajhv5r2etr'

routes = {
    'payment_v2': f'/merchant/v2/{guid}/payment',
    'withdraw_v2': f'/merchant/v2/{guid}/withdraw',
    'payment_query_v2': f'/merchant/v2/{guid}/order/payment',
    'withdraw_query_v2': f'/merchant/v2/{guid}/order/withdraw',
    'balance_v2': f'/merchant/v2/{guid}/balance',
}

def serialize_sign(params, safecode, sign_type):
    sign_fields = []

    if sign_type in ['payment_v2', 'withdraw_v2']:
        sign_fields = [
            'user_id', 'order_id', 'amount', 'currency', 'channel',
            'timestamp',
        ]
    elif sign_type in ['payment_query_v2', 'withdraw_query_v2']:
        sign_fields = [
            'user_id', 'order_id', 'timestamp'
        ]
    elif sign_type in ['balance_v2']:
        sign_fields = ['user_id']
    elif sign_type in ['payment_response']:
        sign_fields = [
            'user_id', 'order_id', 'transaction_id', 'channel',
            'submit_currency', 'submit_amount', 'accept_currency', 'accept_amount',
            'exchange_rate', 'pay_url'
        ]
    elif sign_type in ['withdraw_response']:
        sign_fields = [
            'user_id', 'order_id', 'transaction_id', 'channel',
            'submit_currency', 'submit_amount', 'accept_currency', 'accept_amount',
            'exchange_rate'
        ]
    elif sign_type in ['payment_query_response', 'withdraw_query_response']:
        sign_fields = [
            'user_id', 'order_id', 'transaction_id', 'channel',
            'submit_currency', 'submit_amount', 'accept_currency', 'accept_amount',
            'exchange_rate', 'status', 'timestamp'
        ]
    elif sign_type in ['balance_response']:
        sign_fields = ['user_id']

    if sign_fields:
        # Get only the specified keys from the params dictionary
        params = {key: value for key, value in params.items() if key in sign_fields}

    # Sort the dictionary based on keys
    sorted_params = dict(sorted(params.items()))

    # Concatenate key-value pairs
    sign = '&'.join([f'{key}={value}' for key, value in sorted_params.items()]) + '&' + safecode

    return sign

def generate_signature(data):
    print('data:::' + data)
    with open('key/private_key.pem', 'rb') as key_file:
        private_key = serialization.load_pem_private_key(
            key_file.read(),
            password=None,
        )

    signature = private_key.sign(
        data.encode('utf-8'),
        padding.PKCS1v15(),
        hashes.SHA256()
    )

    encoded_signature = base64.b64encode(signature).decode('utf-8')

    print('base64 encode signature:::' + encoded_signature)

    return encoded_signature

def send_request(request_type, params):    
    # Get Before Sign string
    serialize_sign_string = serialize_sign(params, safecode, request_type)
    print('serialize sign string:::' + serialize_sign_string)
    # Assign sign to params

    #
    sign = generate_signature(serialize_sign_string)
    print('sign base64 encode string:::' + sign)
    params['sign'] = sign

    json_string = json.dumps(params, indent=4, ensure_ascii=False)
    # AES encrypt
    aes = AESCrypt(safecode)
    aes_string = aes.encrypt(json_string).decode('utf-8')
    print('AES string::' + aes_string)
    # Request

    try:
        headers = {'content-type': 'text/plain', 'X-Transaction-Signature': transaction_signature(aes_string, transaction_secret)}
        response = requests.post(url=(url + routes[request_type]), data=aes_string, headers=headers)
        print('response result::' + response.text)

        
    except Exception as e:
        print('Error:', e)
    
    if response.status_code == 200:
        return response.text
    else:
        return None

def verify_signature(pub_key_pem, before_sign_str, sign):
    h = SHA256.new(before_sign_str.encode('utf-8'))
    pubkey = RSA.import_key(pub_key_pem)
    try:
        pkcs1_15.new(pubkey).verify(h, base64.b64decode(sign))
        return True
    except (ValueError, TypeError):
        return False
    
def transaction_signature(request_body, transaction_secret):
    request_body_bytes = bytes(request_body, 'utf-8')
    transaction_secret_bytes = bytes(transaction_secret, 'utf-8')

    hash_object = hmac.new(transaction_secret_bytes, request_body_bytes, hashlib.sha256)

    raw_signature = hash_object.digest()

    signature = base64.b64encode(raw_signature).decode('utf-8')

    return signature

post_params = {
    'user_id': user_id,
    'order_id': f'TEST{user_id}{calendar.timegm(time.gmtime())}',
    'amount': random.randint(10, 1000),
    'currency': 'PHP',
    'channel': 'QRPH',
    'callback_url': 'http://my.service/callback',
    'redirect_url': 'https://my.service/redirect',
    'timestamp': calendar.timegm(time.gmtime()),
    'email': 'aaa@aaa.aaa',
    'remark': 'test'
}

send_request('payment_v2', post_params)